require "defines"

CODE_LINES = 16
CODE_LINES_MAX = 64
ENERGY_USAGE = 200
COMMAND_LIMIT = 10

function copy(obj, seen)

    if type(obj) ~= 'table' then return obj end
    if seen and seen[obj] then return seen[obj] end
    local s = seen or {}
    local res = setmetatable({}, getmetatable(obj))
    s[obj] = res
    for k, v in pairs(obj) do res[copy(k, s)] = copy(v, s) end
    return res

end

function trim(s)

    return s:match "^%s*(.-)%s*$"

end

function glob_init()

    glob["sbox"] = glob["sbox"] or {}
    glob["accumulator"] = glob["accumulator"] or {}
    glob["current"] = glob["current"] or nil
    glob["hashtag"] = glob["hashtag"] or nil

end

function get_hashtag(entity)

    return entity.position.x .. "x" .. entity.position.y

end

function compile_word(s)

    local arg1, arg2, arg3, arg4

    arg1 = s:match "^in:([12345678])$"
    if arg1 then
        return { op = "in", n = tonumber(arg1), block = true }
    end

    arg1 = s:match "^in%?:([12345678])$"
    if arg1 then
        return { op = "in", n = tonumber(arg1), block = false }
    end

    arg1, arg2 = s:match "^out:([12345678]):([lrLR])$"
    if arg1 and arg2 then
        return { op = "out", n = tonumber(arg1), lane = arg2:lower(), condition = "(none)" }
    end

    arg1, arg2, arg3 = s:match "^out%?:([a-zA-Z0-9%-]+):([12345678]):([lrLR])$"
    if arg1 then
        return { op = "out", n = tonumber(arg2), lane = arg3:lower(), condition = arg1 }
    end

    arg1 = s:match "^jump:([1-9][0-9]*)$"
    if arg1 then
        return { op = "jump", n = tonumber(arg1), condtype = "(none)", condvar = "(none)", condval = 0, buffer = "(none)" }
    end

    arg1, arg2, arg3, arg4 = s:match "^jump%?c([=><]):([1-9][0-9]*):(-?%d+):([1-9][0-9]*)$"
    if arg1 and arg2 and arg3 and arg4 then
        return { op = "jump", n = tonumber(arg4), condtype = arg1, condvar = tonumber(arg2), condval = tonumber(arg3), buffer = "(none)" }
    end

    arg1, arg2, arg3, arg4 = s:match "^jump%?c([><]=):([1-9][0-9]*):(-?%d+):([1-9][0-9]*)$"
    if arg1 and arg2 and arg3 and arg4 then
        return { op = "jump", n = tonumber(arg4), condtype = arg1, condvar = tonumber(arg2), condval = tonumber(arg3), buffer = "(none)" }
    end

    arg1, arg2, arg3, arg4 = s:match "^jump%?c(!=):([1-9][0-9]*):(-?%d+):([1-9][0-9]*)$"
    if arg1 and arg2 and arg3 and arg4 then
        return { op = "jump", n = tonumber(arg4), condtype = arg1, condvar = tonumber(arg2), condval = tonumber(arg3), buffer = "(none)" }
    end

    arg1, arg2, arg3 = s:match "^jump%?b(!?=):([a-zA-Z0-9%-]+):([1-9][0-9]*)$"
    if arg1 and arg2 and arg3 then
        return { op = "jump", n = tonumber(arg3), condtype = arg1, condvar = "(none)", condval = 0, buffer = arg2 }
    end

    arg1, arg2 = s:match "^c=:([1-9][0-9]*):(-?%d+)$"
    if arg1 and arg2 then
        return { op = "setcounter", n = tonumber(arg1), v = tonumber(arg2) }
    end

    arg1, arg2 = s:match "^c%+:([1-9][0-9]*):(-?%d+)$"
    if arg1 and arg2 then
        return { op = "inccounter", n = tonumber(arg1), v = tonumber(arg2) }
    end

    arg1, arg2 = s:match "^c%-:([1-9][0-9]*):(-?%d+)$"
    if arg1 and arg2 then
        return { op = "deccounter", n = tonumber(arg1), v = tonumber(arg2) }
    end

    return nil

end

function compile_code(player)

    local i = 0
    local line = ""
    local word = ""
    local token = nil
    local tokens = {}
    local index = {}
    local prev_line = -1
    local cword = nil
    local errors = false

    for i = 1, #glob["current"]["code"] do
        if glob["current"]["code"][i] ~= nil then
            line = glob["current"]["code"][i]
            for word in line:gmatch("%S+") do
                cword = compile_word(word)
                if cword ~= nil then
                    table.insert(tokens, compile_word(word))
                    if prev_line < i then
                        index[i] = #tokens
                        prev_line = i
                    end
                else
                    player.print("error in command '" .. word .. "'")
                    errors = true
                end
            end
        end
    end

    if errors then
        glob["current"]["compiled-code"] = {}
        glob["current"]["line-index"] = {}
    else
        glob["current"]["compiled-code"] = tokens
        glob["current"]["line-index"] = index
    end

    glob["current"]["op"] = 1
 
end

function box_init(entity)

    local hashtag = get_hashtag(entity)
    local pos = entity.position
    local data = {}

    data["buffer"] = "(empty)"
    data["code"] = {}
    data["compiled-code"] = {}
    data["inputs"] = {
        {{ pos.x, pos.y - 2.1 }, { pos.x + 1, pos.y - 2 }},
        {{ pos.x + 2, pos.y - 1 }, { pos.x + 2.1, pos.y }},
        {{ pos.x + 2, pos.y }, { pos.x + 2.1, pos.y + 1 }},
        {{ pos.x, pos.y + 2 }, { pos.x + 1, pos.y + 2.1 }},
        {{ pos.x - 1, pos.y + 2 }, { pos.x, pos.y + 2.1 }},
        {{ pos.x - 2.1, pos.y }, { pos.x - 2, pos.y + 1 }},
        {{ pos.x - 2.1, pos.y - 1 }, { pos.x - 2, pos.y }},
        {{ pos.x - 1, pos.y - 2.1 }, { pos.x, pos.y - 2 }}
    }
    data["outputs"] = {
        l = {
            { x = pos.x + 0.25, y = pos.y - 2.1 },
            { x = pos.x + 2.1, y = pos.y - 0.75 },
            { x = pos.x + 2.1, y = pos.y + 0.25 },
            { x = pos.x + 0.75, y = pos.y + 2.1 },
            { x = pos.x - 0.25, y = pos.y + 2.1 },
            { x = pos.x - 2.1, y = pos.y + 0.75 },
            { x = pos.x - 2.1, y = pos.y - 0.25 },
            { x = pos.x - 0.75, y = pos.y - 2.1 }
        },
        r = {
            { x = pos.x + 0.75, y = pos.y - 2.1 },
            { x = pos.x + 2.1, y = pos.y - 0.25 },
            { x = pos.x + 2.1, y = pos.y + 0.75 },
            { x = pos.x + 0.25, y = pos.y + 2.1 },
            { x = pos.x - 0.75, y = pos.y + 2.1 },
            { x = pos.x - 2.1, y = pos.y + 0.25 },
            { x = pos.x - 2.1, y = pos.y - 0.75 },
            { x = pos.x - 0.25, y = pos.y - 2.1 }
        }
    }
    data["op"] = 1
    data["line-index"] = {}
    data["counter"] = {}

    glob["accumulator"][hashtag] = game.createentity{
        name = "smart-box-accumulator",
        position = entity.position,
        force = entity.force
    }

    glob["sbox"][hashtag] = data

end

function getcounter(hashtag, n)

    local data = glob["sbox"][hashtag]

    if data["counter"][n] == nil then
        data["counter"][n] = 0
    end

    return data["counter"][n]

end

function findopindex(hashtag, n)

    local data = glob["sbox"][hashtag]
    local i = 0

    for i = 1, #data["line-index"] do

        if i >= n and data["line-index"] ~= nil then
            return data["line-index"][i]
        end

    end

    return 1

end

function box_process(hashtag, entity, limit)

    local data = glob["sbox"][hashtag]
    local items = nil
    local accumulator = glob["accumulator"][hashtag]
    local inout = false

    if data["compiled-code"] == nil
            or #data["compiled-code"] == 0
            or accumulator.energy < ENERGY_USAGE
            or limit == 0 then

        return

    end

    accumulator.energy = accumulator.energy - ENERGY_USAGE

    local cmd = data["compiled-code"][data["op"]]

    if cmd.op == "in" then

        if data["buffer"] ~= "(empty)" then
            data["op"] = data["op"] + 1
        else
            items = game.findentitiesfiltered{ area = data["inputs"][cmd.n], name = "item-on-ground" }
            if #items > 0 then
                data["buffer"] = items[1].stack.name
                items[1].destroy()
                data["op"] = data["op"] + 1
            elseif not cmd.block then
                data["op"] = data["op"] + 1
            end
            inout = true
        end

    elseif cmd.op == "out" then

        if data["buffer"] == "(empty)" or (cmd.condition ~= "(none)" and cmd.condition ~= data["buffer"]) then
            data["op"] = data["op"] + 1
        else
            items = game.findentitiesfiltered{ area = data["inputs"][cmd.n], name = "item-on-ground" }
            if #items == 0 then

                inout = true
                game.createentity{
                    name = "item-on-ground",
                    position = data["outputs"][cmd.lane][cmd.n],
                    stack = { name = data["buffer"], count = 1 },
                    force = game.forces.player
                }

                data["buffer"] = "(empty)"
                data["op"] = data["op"] + 1
            end
        end

    elseif cmd.op == "jump" then

        if cmd.condvar == "(none)" and cmd.buffer == "(none)" then
            data["op"] = findopindex(hashtag, cmd.n)
        elseif cmd.condvar == "(none)" then

            if cmd.condtype == "=" and cmd.buffer == data["buffer"]
                    or cmd.condtype == "!=" and cmd.buffer ~= data["buffer"] then

                data["op"] = findopindex(hashtag, cmd.n)

            else
                data["op"] = data["op"] + 1
            end

        else

            countervalue = getcounter(hashtag, cmd.condvar)

            if cmd.condtype == "=" and countervalue == cmd.condval 
                    or cmd.condtype == "!=" and countervalue ~= cmd.condval
                    or cmd.condtype == ">" and countervalue > cmd.condval
                    or cmd.condtype == ">=" and countervalue >= cmd.condval
                    or cmd.condtype == "<" and countervalue < cmd.condval
                    or cmd.condtype == "<=" and countervalue <= cmd.condval then

                data["op"] = findopindex(hashtag, cmd.n)

            else
                data["op"] = data["op"] + 1
            end
        end

    elseif cmd.op == "setcounter" then
        data["counter"][cmd.n] = cmd.v
        data["op"] = data["op"] + 1
    elseif cmd.op == "inccounter" then
        countervalue = getcounter(hashtag, cmd.n)
        data["counter"][cmd.n] = countervalue + cmd.v
        data["op"] = data["op"] + 1
    elseif cmd.op == "deccounter" then
        countervalue = getcounter(hashtag, cmd.n)
        data["counter"][cmd.n] = countervalue - cmd.v
        data["op"] = data["op"] + 1
    end

    if data["op"] > #data["compiled-code"] then
        data["op"] = 1
    end

    if not inout and limit > 1 then
        box_process(hashtag, entity, limit - 1)
    end

end

function gui_open_frame(player, entity) 

    local frame = player.gui.left["smart-box-frame"]

    if frame then
        gui_close_frame(player)
        return
    end

    local hashtag = get_hashtag(entity)
    local i = 0
    local line = ""
    local code_line = nil

    glob["current"] = copy(glob["sbox"][hashtag])
    glob["hashtag"] = hashtag
    glob["scrollpos"] = 1

    frame = player.gui.left.add{
        type = "frame",
        caption = game.localise("smart-box-frame-title"),
        name = "smart-box-frame",
        direction = "vertical"
    }

    local grid = frame.add{
        type = "table",
        colspan = 2,
        name = "smart-box-code-grid"
    }

    for i = 1, CODE_LINES do

        if i <= #glob["current"]["code"] and glob["current"]["code"][i] ~= nil then
            line = glob["current"]["code"][i]
        else
            line = ""
        end

        grid.add{
            type = "label",
            caption = string.format("%02i", i),
            name = "smart-box-line-number-" .. i,
            style = "smart-box-line-number"
        }

        code_line = grid.add{
            type = "textfield",
            text = "",
            name = "smart-box-code-" .. i,
            style = "smart-box-code-line"
        }

        code_line.text = line

    end

    local buttons_grid = frame.add{
        type = "table",
        colspan = 9,
        name = "smart-box-buttons-grid"
    }

    buttons_grid.add{
        type = "button",
        name = "smart-box-save",
        caption = game.localise("smart-box-save"),
        style = "smart-box-small-button"
    }

    buttons_grid.add{
        type = "button",
        name = "smart-box-discard",
        caption = game.localise("smart-box-discard"),
        style = "smart-box-small-button"
    }

    buttons_grid.add{
        type = "label",
        name = "smart-box-separator",
        caption = "--------------"
    }

    buttons_grid.add{
        type = "button",
        name = "smart-box-scroll-top",
        caption = game.localise("smart-box-scroll-top"),
        style = "smart-box-small-button"
    }

    buttons_grid.add{
        type = "button",
        name = "smart-box-scroll-page-up",
        caption = game.localise("smart-box-scroll-page-up"),
        style = "smart-box-small-button"
    }

    buttons_grid.add{
        type = "button",
        name = "smart-box-scroll-up",
        caption = game.localise("smart-box-scroll-up"),
        style = "smart-box-small-button"
    }

    buttons_grid.add{
        type = "button",
        name = "smart-box-scroll-down",
        caption = game.localise("smart-box-scroll-down"),
        style = "smart-box-small-button"
    }

    buttons_grid.add{
        type = "button",
        name = "smart-box-scroll-page-down",
        caption = game.localise("smart-box-scroll-page-down"),
        style = "smart-box-small-button"
    }

    buttons_grid.add{
        type = "button",
        name = "smart-box-scroll-bottom",
        caption = game.localise("smart-box-scroll-bottom"),
        style = "smart-box-small-button"
    }

end

function gui_close_frame(player)

    local frame = player.gui.left["smart-box-frame"]
    frame.destroy()
    glob["current"] = nil
    glob["hashtag"] = nil

end

function gui_save(player)

    local frame = player.gui.left["smart-box-frame"]
    local textfield = nil
    local line = ""
    local i = 0
    local j = 0

    for i = 1, CODE_LINES do
        j = i + glob["scrollpos"] - 1
        textfield = frame["smart-box-code-grid"]["smart-box-code-" .. i]
        line = trim(textfield.text)
        glob["current"]["code"][j] = line
    end

    compile_code(player)

    glob["sbox"][glob["hashtag"]] = copy(glob["current"])
    gui_close_frame(player)

end

function gui_scroll(player, dir)

    local scroll = glob["scrollpos"]
    local frame = player.gui.left["smart-box-frame"]
    local i = 0
    local j = 0
    local line = ""
    local textfield = nil

    for i = 1, CODE_LINES do
        j = i + scroll - 1
        textfield = frame["smart-box-code-grid"]["smart-box-code-" .. i]
        line = textfield.text
        glob["current"]["code"][j] = line
    end

    scroll = scroll + dir

    if scroll < 1 then
        scroll = 1
    end

    if scroll > CODE_LINES_MAX - CODE_LINES + 1 then
        scroll = CODE_LINES_MAX - CODE_LINES + 1
    end

    for i = 1, CODE_LINES do
        j = i + scroll - 1
        if j > #glob["current"]["code"] or glob["current"]["code"][j] == nil then
            line = ""
        else
            line = glob["current"]["code"][j]
        end

        frame["smart-box-code-grid"]["smart-box-code-" .. i].text = line
        frame["smart-box-code-grid"]["smart-box-line-number-" .. i].caption = string.format("%02i", j)

    end

    glob["scrollpos"] = scroll

end

game.oninit(function()

    glob_init()

end)

game.onload(function() 

    glob_init()

end)

game.onevent(defines.events.onbuiltentity, function(event) 

    local player = game.getplayer(event.playerindex)
    local created_entity = event.createdentity
    local entity = nil
    local pos = created_entity.position

    if created_entity.name == "smart-box-programmer" then

        player.insert{name = "smart-box-programmer", count = 1}
        created_entity.destroy()

        entity = game.findentities{{ pos.x, pos.y }, { pos.x, pos.y }}
        if #entity > 0 then
            if entity[1].name == "smart-box" then
                gui_open_frame(player, entity[1])
            end
        end

        return

    end

    if created_entity.name == "smart-box" then

        box_init(created_entity)

    end

end)

game.onevent(defines.events.onentitydied, function(event)

    local entity = event.entity

    if entity.name == "smart-box" then
        local hashtag = get_hashtag(entity.name)
        glob["accumulator"][hashtag].destroy()
        glob["sbox"][hashtag] = nil
    end

end)

game.onevent(defines.events.onguiclick, function(event) 

    local element = event.element
    local player = game.getplayer(event.playerindex)

    if element.name == "smart-box-save" then
        gui_save(player)
        return
    end

    if element.name == "smart-box-discard" then
        gui_close_frame(player)
        return
    end

    if element.name == "smart-box-scroll-top" then
        gui_scroll(player, -CODE_LINES_MAX - 1)
        return
    end

    if element.name == "smart-box-scroll-page-up" then
        gui_scroll(player, -CODE_LINES + 1)
        return
    end

    if element.name == "smart-box-scroll-up" then
        gui_scroll(player, -1)
        return
    end

    if element.name == "smart-box-scroll-down" then
        gui_scroll(player, 1)
        return
    end

    if element.name == "smart-box-scroll-page-down" then
        gui_scroll(player, CODE_LINES - 1)
        return
    end

    if element.name == "smart-box-scroll-bottom" then
        gui_scroll(player, CODE_LINES_MAX + 1)
        return
    end

end)

game.onevent(defines.events.ontick, function(event)

    if game.tick % 3 == 0 then
        for hashtag, entity in pairs(glob["sbox"]) do
            box_process(hashtag, entity, COMMAND_LIMIT)
        end
    end

end)

